import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter_mylibrary/generated/i18n.dart';
import 'package:flutter_mylibrary/http/exception/exception_template.dart';
import 'package:oktoast/oktoast.dart';
import 'package:provider/provider.dart';

import 'view_state.dart';

export 'view_state.dart';

class ViewStateModel with ChangeNotifier {
  /// 防止页面销毁后,异步任务才完成,导致报错
  bool _disposed = false;

  /// 当前的页面状态,默认为busy,可在viewModel的构造方法中指定;
  ViewState _viewState;

  ///context
  BuildContext _appContext;

  BuildContext get appContext => _appContext;

  set appContext(BuildContext context){
    this._appContext = context;
  }

  /// 根据状态构造
  ///
  /// 子类可以在构造函数指定需要的页面状态
  /// FooModel():super(viewState:ViewState.busy);
  ViewStateModel({ViewState viewState})
      : _viewState = viewState ?? ViewState.idle {
    debugPrint('ViewStateModel---constructor--->$runtimeType');
  }


  ///传入对应的ViewModel泛型
  ///[builder] 实例化view model
  ///
  /// 1. 取当前嵌套页面组件VM type数据 [ancestorOfType]=false
  /// 2. 取根页面组件VM type数据 [ancestorOfType]=true&&[builder]==null
  /// 3. 取当前页面存在VM type数据 [ancestorOfType]=false
  static VM of<VM extends ViewStateModel>(BuildContext context,
      {VM Function() builder, String instance, bool ancestorOfType = false}) {
    VM vm;
    _ViewModelStore viewModelStore;
    viewModelStore = _ViewModelStore._of(context);
    vm = viewModelStore?.get<VM>(builder: builder, instanceName: instance);
    if (ancestorOfType) {
      while (viewModelStore != null) {
        vm = viewModelStore.get<VM>(builder: builder, instanceName: instance);
        viewModelStore = _ViewModelStore._of(viewModelStore.context);
      }
    } else {
      while (viewModelStore != null || vm == null) {
        if (viewModelStore != null) {
          vm = viewModelStore.get<VM>(builder: builder, instanceName: instance);
          if (vm != null) break;
          viewModelStore = _ViewModelStore._of(viewModelStore.context);
        }
      }
    }
    assert(vm != null, 'do you forget register the viewModel ${vm.runtimeType} in embed pager or top pager?');
    return vm;
  }



  /// ViewState
  ViewState get viewState => _viewState;

  set viewState(ViewState viewState) {
    _viewStateError = null;
    _viewState = viewState;
    notifyListeners();
  }

  /// ViewStateError
  ViewStateError _viewStateError;

  ViewStateError get viewStateError => _viewStateError;

  /// 以下变量是为了代码书写方便,加入的get方法.严格意义上讲,并不严谨
  ///
  /// get
  bool get isBusy => viewState == ViewState.busy;

  bool get isIdle => viewState == ViewState.idle;

  bool get isEmpty => viewState == ViewState.empty;

  bool get isError => viewState == ViewState.error;

  /// set
  void setIdle() {
    viewState = ViewState.idle;
  }

  void setBusy() {
    viewState = ViewState.busy;
  }

  void setEmpty() {
    viewState = ViewState.empty;
  }

  /// [e]分类Error和Exception两种
  void setError(e, stackTrace, {String message}) {
    ViewStateErrorType errorType = ViewStateErrorType.defaultError;

    /// 见https://github.com/flutterchina/dio/blob/master/README-ZH.md#dioerrortype
    if (e is DioError) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT ||
          e.type == DioErrorType.SEND_TIMEOUT ||
          e.type == DioErrorType.RECEIVE_TIMEOUT) {
        // timeout
        errorType = ViewStateErrorType.networkTimeOutError;
        message = e.error;
      } else if (e.type == DioErrorType.RESPONSE) {
        // incorrect status, such as 404, 503...
        message = e.error;
      } else if (e.type == DioErrorType.CANCEL) {
        // to be continue...
        message = e.error;
      } else {
        // dio将原error重新套了一层
        e = e.error;
        if (e is UnAuthorizedException) {
          stackTrace = null;
          errorType = ViewStateErrorType.unauthorizedError;
        } else if (e is NotSuccessException) {
          stackTrace = null;
          message = e.message;
        } else if (e is SocketException) {
          errorType = ViewStateErrorType.networkTimeOutError;
          message = e.message;
        } else {
          message = e.message;
        }
      }
    }
    viewState = ViewState.error;
    _viewStateError = ViewStateError(
      errorType,
      message: message,
      errorMessage: e.toString(),
    );
    printErrorStack(e, stackTrace);
    onError(viewStateError);
  }

  void onError(ViewStateError viewStateError) {}

  /// 显示错误消息
  showErrorMessage(context, {String message}) {
    if (viewStateError != null || message != null) {
      if (viewStateError.isNetworkTimeOut) {
        message ??= S.of(context).viewStateMessageNetworkError;
      } else {
        message ??= viewStateError.message;
      }
      Future.microtask(() {
        showToast(message, context: context);
      });
    }
  }

  @override
  String toString() {
    return 'BaseModel{_viewState: $viewState, _viewStateError: $_viewStateError}';
  }

  @override
  void notifyListeners() {
    if (!_disposed) {
      super.notifyListeners();
    }
  }

  @override
  void dispose() {
    _disposed = true;
    debugPrint('view_state_model dispose -->$runtimeType');
    super.dispose();
  }
}

/// [e]为错误类型 :可能为 Error , Exception ,String
/// [s]为堆栈信息
printErrorStack(e, s) {
  debugPrint('''
<-----↓↓↓↓↓↓↓↓↓↓-----error-----↓↓↓↓↓↓↓↓↓↓----->
$e
<-----↑↑↑↑↑↑↑↑↑↑-----error-----↑↑↑↑↑↑↑↑↑↑----->''');
  if (s != null) debugPrint('''
<-----↓↓↓↓↓↓↓↓↓↓-----trace-----↓↓↓↓↓↓↓↓↓↓----->
$s
<-----↑↑↑↑↑↑↑↑↑↑-----trace-----↑↑↑↑↑↑↑↑↑↑----->
    ''');
}




///页面组件，实现了当前页面ViewModel装载和卸载
class ViewModelStoreProxyWidget extends StatelessWidget {
  final Widget child;

  const ViewModelStoreProxyWidget({
    this.child,
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MultiProvider(
      providers: [
        Provider<_ViewModelStore>(
          create: (_) => _ViewModelStore(context),
          dispose: (context, value) {
            value._clear();
          },
        )
      ],
      child: child,
    );
  }
}


///内置一个provider管理者
class _ViewModelStore{

  ///用于存储状态
  final _factories = Map<Type, dynamic>();

  /// the ones that get registered by name.
  final _factoriesByName = Map<String, dynamic>();

  BuildContext context;

  _ViewModelStore(this.context);


  ///获取存储provider状态的model
  static _ViewModelStore _of(BuildContext context) {
    try {
      return Provider.of<_ViewModelStore>(context, listen: false);
    } catch (e) {
      debugPrint('当前节点找不到对应ViewModelStore，请确认是否注册ViewModelStoreProxyWidget或context对象是否正确');
      if (kDebugMode) throw e;
    }
  }


  _clear() {
    _factories.values.forEach((value) {
      (value as ViewStateModel).dispose();
    });
    _factoriesByName.values.forEach((value) {
      (value as ViewStateModel).dispose();
    });
    _factories.clear();
    _factoriesByName.clear();
  }


  ///是否注册在provider管理器里
  bool isRegister<T extends ViewStateModel>(T viewModel, {String instanceName}) {
    if (instanceName != null) {
      return _factoriesByName.containsKey(instanceName);
    } else {
      return _factories.containsKey(T);
    }
  }


  ///泛型务必要加上
  ///如果不想使用该方法进行注册view model 请直接使用 [get]
  T register<T extends ViewStateModel>(T viewModel, {String instanceName}) {
    if (instanceName != null) {
      if (_factoriesByName.containsKey(instanceName)) {
        return _factoriesByName[instanceName];
      } else {
        print('register viewModel: $viewModel instanceName: $instanceName');
        viewModel.appContext = context;
        _factoriesByName[instanceName] = viewModel;
        return _factoriesByName[instanceName];
      }
    } else {
      if (_factories.containsKey(T)) {
        return _factories[T];
      } else {
        print('register viewModel: $viewModel instanceName: $instanceName');
        viewModel.appContext = context;
        _factories[T] = viewModel;
        return _factories[T];
      }
    }
  }

  ///不清楚页面是否注册对应 view model 时，需要显示传入一个builder
  ///如果是使用代理穿透，不要穿任何builder，因为会默认在拦截处，实例化一个view model
  T get<T extends ViewStateModel>({T Function() builder, String instanceName}) {
    if (builder != null) {
      if (instanceName != null) {
        return register(builder(), instanceName: instanceName);
      } else {
        return register(builder());
      }
    } else {
      if (instanceName != null) {
        return _factoriesByName[instanceName];
      } else {
        return _factories[T];
      }
    }
  }


}